#####################################################
#                                                   #
#  Source file of the Matrix Elements exports for   #
#  the PwgAmc MG5aMC plugin.                         #
#                                                   #
#####################################################

import os
import logging
import shutil
import itertools
import copy
from math import fmod
import subprocess
import re
import string

plugin_path = os.path.dirname(os.path.realpath( __file__ ))

from madgraph import MadGraph5Error, InvalidCmd, MG5DIR
import madgraph.iolibs.export_fks as export_fks
import madgraph.iolibs.file_writers as writers
import madgraph.various.misc as misc
import madgraph.various.banner as banner_mod
import madgraph.iolibs.helas_call_writers as helas_call_writers
import madgraph.iolibs.files as files
import madgraph.iolibs.drawing_eps as draw
from madgraph.iolibs.files import cp, ln, mv
import madgraph.core.base_objects as base_objects

logger = logging.getLogger('PwgAmc_plugin.MEExporter')

pjoin = os.path.join 

class PwgAmcExporterError(MadGraph5Error):
    """ Error from the Resummation MEs exporter. """ 

class PwgAmcExporter(export_fks.ProcessOptimizedExporterFortranFKS):
    
    # check status of the directory. Remove it if already exists
    check = True 
    # Language type: 'v4' for f77 'cpp' for C++ output
    exporter = 'v4'
    # Output type:
    #[Template/dir/None] copy the Template, just create dir  or do nothing 
    output = 'Template'
    # Decide which type of merging if used [madevent/madweight]
    grouped_mode = False
    # if no grouping on can decide to merge uu~ and u~u anyway:
    sa_symmetry = False

    template_path = pjoin(plugin_path,'mg5amcpwgTemplate')
    
    def __init__(self, *args, **opts):
        """ Possibly define extra instance attribute for this daughter class."""
        return super(PwgAmcExporter, self).__init__(*args, **opts)
    
    def read_template_file(self, name):
        """ Read a given template file. In conjunction of the use of class attributes,
        this is to help making the choice of the template file modular."""
        return open(pjoin(self.template_path,name),'r').read()
    
    def copy_fkstemplate(self, *args, **opts):
        """Additional actions needed for setup of Template
        """

        dir_path = self.dir_path
        clean =self.opt['clean']
        
        #First copy the full template tree if dir_path doesn't exit
        if not os.path.isdir(dir_path):
            logger.info('initialize a new directory: %s' % \
                        os.path.basename(dir_path))
            shutil.copytree(os.path.join(self.template_path, 'TEMPLATE'), dir_path, True)
        
        # We must link the CutTools to the Library folder of the active Template
        self.link_CutTools(dir_path)
        # We must link the TIR to the Library folder of the active Template
        link_tir_libs=[]
        tir_libs=[]
        tir_include=[]
        for tir in self.all_tir:
            tir_dir="%s_dir"%tir
            libpath=getattr(self,tir_dir)
            libpath = self.link_TIR(os.path.join(self.dir_path, 'lib'),
                                              libpath,"lib%s.a"%tir,tir_name=tir)
            setattr(self,tir_dir,libpath)
            if libpath != "":
                if tir in ['pjfry','ninja','golem', 'samurai','collier']:
                    # We should link dynamically when possible, so we use the original
                    # location of these libraries.
                    link_tir_libs.append('-L%s/ -l%s'%(libpath,tir))
                    tir_libs.append('%s/lib%s.$(libext)'%(libpath,tir))
                    # For Ninja, we must also link against OneLoop.
                    if tir in ['ninja']:
                        if not any(os.path.isfile(pjoin(libpath,'libavh_olo.%s'%ext)) 
                                              for ext in ['a','dylib','so']):
                            raise MadGraph5Error(
"The OneLOop library 'libavh_olo.(a|dylib|so)' could no be found in path '%s'. Please place a symlink to it there."%libpath)
                        link_tir_libs.append('-L%s/ -l%s'%(libpath,'avh_olo'))
                        tir_libs.append('%s/lib%s.$(libext)'%(libpath,'avh_olo'))
                    # We must add the corresponding includes for these TIR
                    if tir in ['golem','samurai','ninja','collier']:
                        trg_path = pjoin(os.path.dirname(libpath),'include')
                        if os.path.isdir(trg_path):
                            to_include = misc.find_includes_path(trg_path,
                                                        self.include_names[tir])
                        else:
                            to_include = None
                        # Special possible location for collier
                        if to_include is None and tir=='collier':
                            to_include = misc.find_includes_path(
                               pjoin(libpath,'modules'),self.include_names[tir])
                        if to_include is None:
                            logger.error(
'Could not find the include directory for %s, looking in %s.\n' % (tir ,str(trg_path))+
'Generation carries on but you will need to edit the include path by hand in the makefiles.')
                            to_include = '<Not_found_define_it_yourself>'
                        tir_include.append('-I %s'%to_include)
                else:
                    link_tir_libs.append('-l%s'%tir)
                    tir_libs.append('$(LIBDIR)lib%s.$(libext)'%tir)
        
        dirpath = os.path.join(self.dir_path, 'SubProcesses')
        filename = pjoin(self.dir_path, 'SubProcesses','makefile_loop')
        calls = self.write_makefile_TIR(writers.MakefileWriter(filename),
                                    link_tir_libs,tir_libs,tir_include=tir_include)
        os.remove(os.path.join(self.dir_path,'Source','make_opts.inc'))
        filename = pjoin(self.dir_path, 'Source','make_opts')
        calls = self.write_make_opts(writers.MakefileWriter(filename),
                                                        link_tir_libs,tir_libs)
        
        # Duplicate run_card and FO_analyse_card
        for card in ['FO_analyse_card', 'shower_card']:
            try:
                shutil.copy(pjoin(self.dir_path, 'Cards',
                                         card + '.dat'),
                           pjoin(self.dir_path, 'Cards',
                                        card + '_default.dat'))
            except IOError:
                logger.warning("Failed to copy " + card + ".dat to default")

        cwd = os.getcwd()
        dirpath = os.path.join(self.dir_path, 'SubProcesses')
        try:
            os.chdir(dirpath)
        except os.error:
            logger.error('Could not cd to directory %s' % dirpath)
            return 0

        # We add here the user-friendly MadLoop option setter.
        cpfiles= ["SubProcesses/MadLoopParamReader.f",
                  "Cards/MadLoopParams.dat",
                  "SubProcesses/MadLoopParams.inc"]
        
        for file in cpfiles:
            shutil.copy(os.path.join(self.loop_dir,'StandAlone/', file),
                        os.path.join(self.dir_path, file))
        
        shutil.copy(pjoin(self.dir_path, 'Cards','MadLoopParams.dat'),
                    pjoin(self.dir_path, 'Cards','MadLoopParams_default.dat'))

        if os.path.exists(pjoin(self.dir_path, 'Cards', 'MadLoopParams.dat')):          
                self.MadLoopparam = banner_mod.MadLoopParam(pjoin(self.dir_path, 
                                                  'Cards', 'MadLoopParams.dat'))
                # write the output file
                self.MadLoopparam.write(pjoin(self.dir_path,"SubProcesses",
                                                           "MadLoopParams.dat"))
                                       
        # We need minimal editing of MadLoopCommons.f
        MadLoopCommon = open(os.path.join(self.loop_dir,'StandAlone', 
                                    "SubProcesses","MadLoopCommons.inc")).read()
        writer = writers.FortranWriter(os.path.join(self.dir_path, 
                                             "SubProcesses","MadLoopCommons.f"))
        writer.writelines(MadLoopCommon%{
                                   'print_banner_commands':self.MadLoop_banner},
                                            context={'collier_available':self.tir_available_dict['collier']})
        writer.close()
                                       
        # Write the cts_mpc.h and cts_mprec.h files imported from CutTools
        self.write_mp_files(writers.FortranWriter('cts_mprec.h'),\
                                           writers.FortranWriter('cts_mpc.h'))

        
        # Finally make sure to turn off MC over Hel for the default mode.
        FKS_card_path = pjoin(self.dir_path,'Cards','FKS_params.dat')
        FKS_card_file = open(FKS_card_path,'r')
        FKS_card = FKS_card_file.read()
        FKS_card_file.close()
        FKS_card = re.sub(r"#NHelForMCoverHels\n-?\d+",
                                             "#NHelForMCoverHels\n-1", FKS_card)
        FKS_card_file = open(FKS_card_path,'w')
        FKS_card_file.write(FKS_card)
        FKS_card_file.close()

        # Return to original PWD
        os.chdir(cwd)

        # MZMZ python files are not needed at the moment
        # Copy the different python files in the Template
        self.copy_python_files()

        # We need to create the correct open_data for the pdf
        self.write_pdf_opendata()



    def generate_pwgamc_files(self, matrix_elements, fortran_model, 
                                    path=os.getcwd(), OLP='MadLoop'):
        """ here we do not need any more the usual directory structure (P0*),
        all files will be put into the "SubProcesses" directory. 
        matrix_elements is a FKSHelasMultiProcess
        """
        self.output_dir = pjoin(self.dir_path, "SubProcesses")
        self.fortran_model = fortran_model

        # model
        if not self.model:
            self.model = matrix_elements['matrix_elements'][0].born_matrix_element.get('processes')[0].get('model')

        if not os.path.isdir(self.output_dir):
            os.mkdir(self.output_dir)
        
        cwd = os.getcwd()
        os.chdir(self.output_dir)

        calls, born_processes, born_pdgs = self.generate_born_files(matrix_elements['matrix_elements'])
        real_calls, real_pdgs = self.generate_real_files(matrix_elements)
        calls += real_calls
        calls+= self.generate_loop_files(matrix_elements['matrix_elements'])

        self.generate_global_files(matrix_elements)

        # now write init_processes above SubProcesses
        os.chdir(self.dir_path)
        filename = "init_processes.f"
        nborn, nreal = self.write_init_processes(writers.FortranWriter(filename), \
                                 born_pdgs, real_pdgs)

        filename = "nlegborn.h"
        self.write_nlegborn(writers.FortranWriter(filename), \
                                 nborn, nreal)

        # finally go bach to the original wd to avoid side effects
        os.chdir(cwd)

        return calls, born_processes

    
    def generate_born_files(self, matrix_elements):
        """ write the born files and all the needed wrappers
        """

        born_processes = []
        pdg_list = []
        calls = 0
        maxcamps = 0
        maxgraphs = 0

        # loop over the various born processes
        for ime, me in enumerate(matrix_elements):
            born_processes.append(me.get('processes'))

            pdgs = [[l['id'] for l in p.get('legs')] for p in me.get('processes')]
            pdg_list.append(pdgs)

            # write the file with the born matrix element
            tag = self.get_born_tag(ime+1)
            filename = "born_%s.f" % tag
            calls_born, ncolor_born, ndiags_born = \
                self.write_born_pwgamc(writers.FortranWriter(filename),\
                                       me, tag,\
                                       self.fortran_model)
            calls += calls_born
            maxcamps = max(maxcamps, ncolor_born)
            maxgraphs = max(maxgraphs, ndiags_born)

            # draw the feynman diagrams
            filename = "born_%s.ps" % tag
            plot = draw.MultiEpsDiagramDrawer(me.born_matrix_element.\
                            get('base_amplitude').get('diagrams'),
                            filename,
                            model=me.born_matrix_element.\
                            get('processes')[0].get('model'),
                            amplitude=True, diagram_type='born')
            plot.draw()
    
            # write the color links
            for i in range(len(me.color_links)):
                filename = 'b_sf_%s_%3.3d.f' % (tag, i + 1)              
                self.write_b_sf_pwgamc(writers.FortranWriter(filename),
                         me, i, tag,
                         self.fortran_model)

            # and their wrapper function
            filename = 'sborn_sf_%s.f' % tag              
            self.write_sborn_sf_pwgamc(writers.FortranWriter(filename),
                         me.color_links, tag,
                         self.fortran_model)

            # write the leshouches file with color informations
            filename = 'leshouches_%s.inc' % tag
            nflows = self.write_leshouches_file(writers.FortranWriter(filename), me.born_matrix_element)

            filename = 'born_nhel_%s.inc' % tag
            self.write_born_nhel_file(writers.FortranWriter(filename),
                               me.born_matrix_element, nflows,
                               self.fortran_model,
                               ncolor_born)

        # include file with the maximum number of color amplitudes
        filename = 'maxamps_B.inc'
        self.write_maxamps_file(writers.FortranWriter(filename), maxcamps, maxgraphs)

        # wrapper on born processes, it uses the pdgs as stored in the loop
        # for the botn MEs
        filename = 'sborn_proc.f'
        self.write_born_wrapper(writers.FortranWriter(filename), pdg_list)

        # write the list of processes (pdgs)
        filename = "proc_label"
        self.write_proc_label(filename, pdg_list)

        return calls, born_processes, pdg_list


    def get_sorted_pdg_list(self, pdg_list, nincoming):
        """return a list where the final state pdgs are sorted
        """
        return pdg_list[:nincoming] + sorted(pdg_list[nincoming:])

    
    def generate_real_files(self, matrix_elements):
        """ write the real-emission files and all the needed wrappers
        matrix_elements is a list of real matrix elements
        """
        calls = 0
        pdg_list = []
        maxcamps = 0
        maxgraphs = 0
        sorted_pdg_list = []

        next_nin_list = []

        # loop over the various real processes
        for born in matrix_elements['matrix_elements']:
            for real in born.real_processes:
                nexternal, ninitial = real.matrix_element.get_nexternal_ninitial()
                # MZMZ this may not be fully optimal, check it 
                pdgs = [[l['id'] for l in p.get('legs')] for p in real.matrix_element.get('processes')]
                if any([pdgs[0] in ppdg for ppdg in pdg_list]):
                    # nothing to do here, we have already this real emission
                    continue
                if self.get_sorted_pdg_list(pdgs[0], ninitial) in sorted_pdg_list: 
                    # it can be that the same real exists with differen leg numbers
                    continue

                pdg_list.append(pdgs)
                sorted_pdg_list.extend([self.get_sorted_pdg_list(pdg, ninitial) for pdg in pdgs])

                # write the file with the real matrix element
                tag = self.get_real_tag(len(pdg_list))
                filename = "real_%s.f" % tag
                calls_real, ncolor_real, ndiags_real = \
                    self.write_real_matrix_element_pwgamc(writers.FortranWriter(filename), 
                                            real.matrix_element, tag,
                                            self.fortran_model)
                calls += calls_real
                maxcamps = max(maxcamps, ncolor_real)
                maxgraphs = max(maxgraphs, ndiags_real)

                next_nin_list.append(real.matrix_element.get_nexternal_ninitial())

                # draw the feynman diagrams
                filename = "real_%s.ps" % tag
                plot = draw.MultiEpsDiagramDrawer(real.matrix_element.\
                                get('base_amplitude').get('diagrams'),
                                filename,
                                model=real.matrix_element.\
                                get('processes')[0].get('model'),
                                amplitude=True, diagram_type='born')
                plot.draw()

                # write the leshouches file with color informations
                filename = 'leshouches_%s.inc' % tag
                self.write_leshouches_file(writers.FortranWriter(filename), real.matrix_element)

        # check that a single nexternal file is needed
        next_nin_list = list(set(next_nin_list))
        if len(next_nin_list) != 1:
            raise PwgAmcExporterError("output of processes with different number" + \
                    " of external/initial legs not supported (found %d values)" % len(next_nin_list))
        nexternal, ninitial = next_nin_list[0]
        filename = 'nexternal.inc'
        self.write_nexternal_file(writers.FortranWriter(filename),
                             nexternal, ninitial)

        # include file with the maximum number of color amplitudes
        filename = 'maxamps_R.inc'
        self.write_maxamps_file(writers.FortranWriter(filename), maxcamps, maxgraphs)

        # wrapper on real emission, it uses the pdgs as stored in the loop
        # for the real emission MEs
        filename = 'sreal_proc.f'
        self.write_real_wrapper(writers.FortranWriter(filename), pdg_list)

        # write the list of processes (pdgs)
        filename = "proc_label_real"
        self.write_proc_label(filename, pdg_list)

        return calls, pdg_list



    def generate_loop_files(self, matrix_elements):
        """ write the virtual (loop) files and all the needed wrappers
        """

        pdg_list = []

        self.madloop_makefile_name = 'makefile_loop'

        os.mkdir(os.path.join(self.dir_path,'SubProcesses','MadLoop5_resources'))

        # loop over the various virtual processes
        for ime, me in enumerate(matrix_elements):
            pdgs = [[l['id'] for l in p.get('legs')] \
                    for p in me.virt_matrix_element.get('processes')]
            pdg_list.append(pdgs)

            self.virt_tag = self.get_virt_tag(ime+1)
            self.generate_loop_subprocess(me.virt_matrix_element, 
                                         self.fortran_model, 
                                         unique_id=ime+1)

        # wrapper on born processes, it uses the pdgs as stored in the loop
        # for the botn MEs
        filename = 'svirt_proc.f'
        self.write_virt_wrapper(writers.FortranWriter(filename), pdg_list)

        # write global_specs
        self.write_global_specs([me.virt_matrix_element for me in matrix_elements], \
                output_path=pjoin(self.dir_path,'SubProcesses','global_specs.inc'))

        return 0



    def generate_global_files(self, matrix_elements):
        """ write global files 
        """
        pass


    def get_ME_identifier(self, matrix_element, *args, **opts):
        """ A function returning a string uniquely identifying the matrix 
        element given in argument so that it can be used as a prefix to all
        MadLoop5 subroutines and common blocks related to it. This allows
        to compile several processes into one library as requested by the 
        BLHA (Binoth LesHouches Accord) guidelines. Returns the virt_tag
        """
     
        return "%s_" % self.virt_tag


    # writer functions
    #===============================================================================
    # write_maxamps_file
    #===============================================================================
    def write_maxamps_file(self, writer, maxamps, maxgraphs):
        """write the maxamps file"""
        content = "integer maxamps,maxgraphs\nparameter (maxamps=%d)\nparameter (maxgraphs=%d)\n" \
                % (maxamps, maxgraphs)
        writer.writelines(content)


    # writer functions
    #===============================================================================
    # write_leshouche_file
    #===============================================================================
    def write_leshouches_file(self, writer, matrix_element):
        """write the maxamps file"""
        lines = []

        # Extract number of external particles
        (nexternal, ninitial) = matrix_element.get_nexternal_ninitial()

        for iproc, proc in enumerate(matrix_element.get('processes')):
            legs = proc.get_legs_with_decays()

            # Here goes the color connections corresponding to the JAMPs
            # Only one output, for the first subproc!
            if iproc == 0:
                # If no color basis, just output trivial color flow
                if not matrix_element.get('color_basis'):
                    for i in [1, 2]:
                        for ileg in range(nexternal):
                            lines.append('ICOLUP(%d, %d, 1) = 0' % (i, ileg+1))

                else:
                    # First build a color representation dictionnary
                    repr_dict = {}
                    for l in legs:
                        repr_dict[l.get('number')] = \
                            proc.get('model').get_particle(l.get('id')).get_color()\
                            * (-1)**(1+l.get('state'))
                    # Get the list of color flows
                    color_flow_list = \
                        matrix_element.get('color_basis').color_flow_decomposition(repr_dict,
                                                                                   ninitial)
                    # And output them properly
                    for cf_i, color_flow_dict in enumerate(color_flow_list):
                        for i in [0, 1]:
                            for ileg in range(nexternal): 
                                lines.append('ICOLUP(%d, %d, %d) = %3r' % \
                                        (i+1, ileg+1, cf_i+1, color_flow_dict[legs[ileg].get('number')][i]))

        writer.writelines(lines)

        return len(color_flow_list)

    #===============================================================================
    # write_nleg_born
    #===============================================================================
    def write_nlegborn(self, writer, nborn, nreal):
        """write the init_processes file"""

        replace_dict = {'maxprocborn': nborn, 'maxprocreal': nreal}

        file = open(os.path.join(self.template_path, \
                          'nlegborn.inc')).read()
        file = file % replace_dict
        
        # Write the file
        writer.writelines(file)


    # writer functions
    #===============================================================================
    # write_init_processes
    #===============================================================================
    def write_init_processes(self, writer, born_pdg_list, real_pdg_list):
        """write the init_processes file"""

        replace_dict = {}
        replace_dict['flst_nborn'] = len(sum(born_pdg_list, []))
        replace_dict['flst_nreal'] = len(sum(real_pdg_list, []))

        born_lines = ""
        for ib, born_pdg in enumerate(sum(born_pdg_list, [])):
            for ipdg, pdg in enumerate(born_pdg):
                if pdg == 21:
                    pdg = 0
                born_lines += "flst_born(%d, %d) = %d\n" % (ipdg + 1, ib + 1, pdg)
            born_lines += '\n'

        real_lines = ""
        for ir, real_pdg in enumerate(sum(real_pdg_list, [])):
            for ipdg, pdg in enumerate(real_pdg):
                if pdg == 21:
                    pdg = 0
                real_lines += "flst_real(%d, %d) = %d\n" % (ipdg + 1, ir + 1, pdg)
            real_lines += '\n'

        replace_dict['flst_born_lines'] = born_lines
        replace_dict['flst_real_lines'] = real_lines

        file = open(os.path.join(self.template_path, \
                          'init_processes.inc')).read()
        file = file % replace_dict
        
        # Write the file
        writer.writelines(file)

        return replace_dict['flst_nborn'], replace_dict['flst_nreal']


    #===============================================================================
    # write_born_wrapper
    #===============================================================================
    def write_born_wrapper(self, writer, pdg_list):
        """write the maxamps file"""

        iflinesME_template = "if (%s) then\ncall sborn_cl_%s(p_born1,wgtmunu1,wgt1)\ncall sborn_sf_%s(p_born1,wgtjk1)\ngoto 20\n" 
        commonlinesAMP_template = "double Precision amp2_%d(maxgraphs), jamp2_%d(0:maxamps)\ncommon/to_amps_%s/amp2_%d, jamp2_%d\n"
        iflinesLH_template = "if (%s) then\ninclude \"leshouches_%s.inc\"\n" + \
                         "call store_jamp2cum(jamp2cum, jamp2_%d, iflow)\n goto 20\n"

        replace_dict = {'iflines_ME': '', 'commonlines_AMP': '', 'iflines_LH': ''}


        for i, pdgs in enumerate(pdg_list):
            tag = self.get_born_tag(i+1) 
            pdgstring = self.get_iflines_pdgs(pdgs)
            
            if (i == 0):
                replace_dict['iflines_ME'] += iflinesME_template % (pdgstring, tag, tag)
                replace_dict['iflines_LH'] += iflinesLH_template % (pdgstring, tag, i+1)
            else:
                replace_dict['iflines_ME'] += ('else'+iflinesME_template) % (pdgstring, tag, tag)
                replace_dict['iflines_LH'] += ('else'+iflinesLH_template) % (pdgstring, tag, i+1)
            
            replace_dict['commonlines_AMP'] += commonlinesAMP_template % (i+1, i+1, tag, i+1, i+1)

        replace_dict['iflines_ME'] += 'endif\n'
        replace_dict['iflines_LH'] += 'endif\n'

        file = open(os.path.join(self.template_path, \
                          'sborn_proc_pwgamc.inc')).read()
        file = file % replace_dict
        
        # Write the file
        writer.writelines(file)


    # writer functions
    #===============================================================================
    # write_real_wrapper
    #===============================================================================
    def write_real_wrapper(self, writer, pdg_list):
        """write the maxamps file"""

        iflinesME_template = "if (%s) then\ncall srealmtrx_%s(p1,wgt)\ngoto 20\n" 
        commonlinesAMP_template = "double Precision jamp2_%d(0:maxamps)\ncommon/to_amps_%s/jamp2_%d\n"
        iflinesLH_template = "if (%s) then\ninclude \"leshouches_%s.inc\"\n" + \
                         "call store_jamp2cum(jamp2cum, jamp2_%d, iflow)\n goto 20\n"

        replace_dict = {'iflines_ME': '', 'commonlines_AMP': '', 'iflines_LH': ''}


        for i, pdgs in enumerate(pdg_list):
            tag = self.get_real_tag(i+1) 
            pdgstring = self.get_iflines_pdgs(pdgs)
            
            if (i == 0):
                replace_dict['iflines_ME'] += iflinesME_template % (pdgstring, tag)
                replace_dict['iflines_LH'] += iflinesLH_template % (pdgstring, tag, i+1)
            else:
                replace_dict['iflines_ME'] += ('else'+iflinesME_template) % (pdgstring, tag)
                replace_dict['iflines_LH'] += ('else'+iflinesLH_template) % (pdgstring, tag, i+1)
            
            replace_dict['commonlines_AMP'] += commonlinesAMP_template % (i+1, tag, i+1)

        replace_dict['iflines_ME'] += 'endif\n'
        replace_dict['iflines_LH'] += 'endif\n'

        file = open(os.path.join(self.template_path, \
                          'sreal_proc_pwgamc.inc')).read()
        file = file % replace_dict
        
        # Write the file
        writer.writelines(file)


    #===============================================================================
    # write_virt_wrapper
    #===============================================================================
    def write_virt_wrapper(self, writer, pdg_list):
        """write the maxamps file"""

        iflinesME_template = "if (%s) then\ncall %s_SLOOPMATRIX(p_born1,virt_wgts)\ngoto 20\n" 

        replace_dict = {'iflines_ME': ''}

        for i, pdgs in enumerate(pdg_list):
            tag = self.get_virt_tag(i+1) 
            pdgstring = self.get_iflines_pdgs(pdgs)
            
            if (i == 0):
                replace_dict['iflines_ME'] += iflinesME_template % (pdgstring, tag)
            else:
                replace_dict['iflines_ME'] += ('else'+iflinesME_template) % (pdgstring, tag)
            
        replace_dict['iflines_ME'] += 'endif\n'

        file = open(os.path.join(self.template_path, \
                          'svirt_proc_pwgamc.inc')).read()
        file = file % replace_dict
        
        # Write the file
        writer.writelines(file)






    # writer functions
    #===============================================================================
    # write_born_pwgamc
    #===============================================================================
    def write_born_pwgamc(self, writer, fksborn, metag, fortran_model):
        """Export a matrix element to a born.f file in MadFKS format"""

        matrix_element = fksborn.born_matrix_element
        
        if not matrix_element.get('processes') or \
               not matrix_element.get('diagrams'):
            return 0
    
        if not isinstance(writer, writers.FortranWriter):
            raise writers.FortranWriter.FortranWriterError(\
                "writer not FortranWriter")
        # Set lowercase/uppercase Fortran code
        writers.FortranWriter.downcase = False
    
        replace_dict = {}

        # a tag for the matrix element
        replace_dict['metag'] = metag
    
        # Extract version number and date from VERSION file
        info_lines = self.get_mg5_info_lines()
        replace_dict['info_lines'] = info_lines
    
        # Extract process info lines
        process_lines = self.get_process_info_lines(matrix_element)
        replace_dict['process_lines'] = process_lines
    
        # Extract ncomb
        ncomb = matrix_element.get_helicity_combinations()
        replace_dict['ncomb'] = ncomb
    
        # Extract helicity lines
        helicity_lines = self.get_helicity_lines(matrix_element)
        replace_dict['helicity_lines'] = helicity_lines
    
        # Extract IC line
        ic_line = self.get_ic_line(matrix_element)
        replace_dict['ic_line'] = ic_line
    
        # Extract overall denominator
        # Averaging initial state color, spin, and identical FS particles
        den_factor_line = self.get_den_factor_line(matrix_element)
        replace_dict['den_factor_line'] = den_factor_line
    
        # Extract ngraphs
        ngraphs = matrix_element.get_number_of_amplitudes()
        replace_dict['ngraphs'] = ngraphs
    
        # Extract nwavefuncs
        nwavefuncs = matrix_element.get_number_of_wavefunctions()
        replace_dict['nwavefuncs'] = nwavefuncs
    
        # Extract ncolor
        ncolor = max(1, len(matrix_element.get('color_basis')))
        replace_dict['ncolor'] = ncolor
    
        # Extract color data lines
        color_data_lines = self.get_color_data_lines(matrix_element)
        replace_dict['color_data_lines'] = "\n".join(color_data_lines)
    
        # Extract helas calls
        helas_calls = fortran_model.get_matrix_element_calls(\
                    matrix_element)
        replace_dict['helas_calls'] = "\n".join(helas_calls)
    
        # Extract amp2 lines
        amp2_lines = self.get_amp2_lines(matrix_element)
        replace_dict['amp2_lines'] = '\n'.join(amp2_lines)
    
        # Extract JAMP lines
        jamp_lines = self.get_JAMP_lines(matrix_element)
        replace_dict['jamp_lines'] = '\n'.join(jamp_lines)

        # Set the size of Wavefunction
        if not self.model or any([p.get('spin') in [4,5] for p in self.model.get('particles') if p]):
            replace_dict['wavefunctionsize'] = 20
        else:
            replace_dict['wavefunctionsize'] = 8

        # Extract glu_ij_lines
        ij_lines = self.get_ij_lines(fksborn)
        replace_dict['ij_lines'] = '\n'.join(ij_lines)

        # Extract born_pdg
        bpdg_lines = self.get_born_pdg_lines(fksborn)
        replace_dict['born_pdg_lines'] = '\n'.join(bpdg_lines)

        # Extract the number of FKS process
        replace_dict['nconfs'] = max(len(fksborn.get_fks_info_list()),1)

        file = open(os.path.join(self.template_path, \
                          'born_pwgamc.inc')).read()
        file = file % replace_dict
        
        # Write the file
        writer.writelines(file)
    
        return len([call for call in helas_calls if call.find('#') != 0]), ncolor, ngraphs
    

    
    #===============================================================================
    # write_b_sf_pwgamc
    #===============================================================================
    #test written
    def write_b_sf_pwgamc(self, writer, fksborn, i, metag, fortran_model):
        """Create the b_sf_xxx.f file for the soft linked born in MadFKS format"""

        matrix_element = copy.copy(fksborn.born_matrix_element)

        if not matrix_element.get('processes') or \
               not matrix_element.get('diagrams'):
            return 0
    
        if not isinstance(writer, writers.FortranWriter):
            raise writers.FortranWriter.FortranWriterError(\
                "writer not FortranWriter")
        # Set lowercase/uppercase Fortran code
        writers.FortranWriter.downcase = False

        iborn = i + 1
        link = fksborn.color_links[i]
    
        replace_dict = {}

        # a tag for the matrix element
        replace_dict['metag'] = metag
        
        # number of color link
        replace_dict['iborn'] = iborn
    
        # Extract version number and date from VERSION file
        info_lines = self.get_mg5_info_lines()
        replace_dict['info_lines'] = info_lines 
    
        # Extract process info lines
        process_lines = self.get_process_info_lines(matrix_element)
        replace_dict['process_lines'] = process_lines + \
            "\nc spectators: %d %d \n" % tuple(link['link'])
    
        # Extract ncomb
        ncomb = matrix_element.get_helicity_combinations()
        replace_dict['ncomb'] = ncomb
    
        # Extract helicity lines
        helicity_lines = self.get_helicity_lines(matrix_element)
        replace_dict['helicity_lines'] = helicity_lines
    
        # Extract IC line
        ic_line = self.get_ic_line(matrix_element)
        replace_dict['ic_line'] = ic_line

        # Extract overall denominator
        # Averaging initial state color, spin, and identical FS particles
        den_factor_line = self.get_den_factor_line(matrix_element)
        replace_dict['den_factor_line'] = den_factor_line

        # Extract ngraphs
        ngraphs = matrix_element.get_number_of_amplitudes()
        replace_dict['ngraphs'] = ngraphs
    
        # Extract nwavefuncs
        nwavefuncs = matrix_element.get_number_of_wavefunctions()
        replace_dict['nwavefuncs'] = nwavefuncs

    
        # Extract ncolor
        ncolor1 = max(1, len(link['orig_basis']))
        replace_dict['ncolor1'] = ncolor1
        ncolor2 = max(1, len(link['link_basis']))
        replace_dict['ncolor2'] = ncolor2
    
        # Extract color data lines
        color_data_lines = self.get_color_data_lines_from_color_matrix(\
                                link['link_matrix'])
        replace_dict['color_data_lines'] = "\n".join(color_data_lines)
    
        # Extract amp2 lines
        amp2_lines = self.get_amp2_lines(matrix_element)
        replace_dict['amp2_lines'] = '\n'.join(amp2_lines)
    
        # Extract JAMP lines
        jamp_lines = self.get_JAMP_lines(matrix_element)
        new_jamp_lines = []
        for line in jamp_lines:
            line = line.replace('JAMP', 'JAMP1')
            new_jamp_lines.append(line)
        replace_dict['jamp1_lines'] = '\n'.join(new_jamp_lines)
    
        matrix_element.set('color_basis', link['link_basis'] )
        jamp_lines = self.get_JAMP_lines(matrix_element)
        new_jamp_lines = []
        for line in jamp_lines:
            line = line.replace('JAMP', 'JAMP2')
            new_jamp_lines.append(line)
        replace_dict['jamp2_lines'] = '\n'.join(new_jamp_lines)
    
    
        # Extract the number of FKS process
        replace_dict['nconfs'] = len(fksborn.get_fks_info_list())

        file = open(os.path.join(self.template_path, \
                          'b_sf_xxx_pwgamc.inc')).read()
        file = file % replace_dict
        
        # Write the file
        writer.writelines(file)
    
        return 0 , ncolor1


    # writer functions
    #===============================================================================
    # write_sborn_sf_pwgamc
    #===============================================================================
    def write_sborn_sf_pwgamc(self, writer, color_links, metag, fortran_model):
        """Export a matrix element to a born.f file in MadFKS format"""

    
        if not isinstance(writer, writers.FortranWriter):
            raise writers.FortranWriter.FortranWriterError(\
                "writer not FortranWriter")
        # Set lowercase/uppercase Fortran code
        writers.FortranWriter.downcase = False
    
        replace_dict = {}

        # a tag for the matrix element
        replace_dict['metag'] = metag
        replace_dict['sb_sf_lines'] = ''

        for (i, c_link) in enumerate(color_links):
            m, n = c_link['link']
            replace_dict['sb_sf_lines'] += "call sb_sf_%s_%3.3d(p, wgtjk(%d,%d))\n" % (metag, i+1, m, n)
            replace_dict['sb_sf_lines'] += "wgtjk(%d,%d)=wgtjk(%d,%d)\n\n" % (n, m, m, n)
    
        file = open(os.path.join(self.template_path, \
                          'sborn_sf_pwgamc.inc')).read()
        file = file % replace_dict
        
        # Write the file
        writer.writelines(file)
    



    #===============================================================================
    # write_real_matrix_element_pwgamc
    #===============================================================================
    #test written
    def write_real_matrix_element_pwgamc(self, writer, matrix_element, metag, fortran_model):
        """Export a matrix element to a matrix.f file in MG4 madevent format"""
    
        if not matrix_element.get('processes') or \
               not matrix_element.get('diagrams'):
            return 0,0
    
        if not isinstance(writer, writers.FortranWriter):
            raise writers.FortranWriter.FortranWriterError(\
                "writer not FortranWriter")
        # Set lowercase/uppercase Fortran code
        writers.FortranWriter.downcase = False
    
        replace_dict = {}
        replace_dict['metag'] = metag
    
        # Extract version number and date from VERSION file
        info_lines = self.get_mg5_info_lines()
        replace_dict['info_lines'] = info_lines
    
        # Extract process info lines
        process_lines = self.get_process_info_lines(matrix_element)
        replace_dict['process_lines'] = process_lines
    
        # Extract ncomb
        ncomb = matrix_element.get_helicity_combinations()
        replace_dict['ncomb'] = ncomb
    
        # Extract helicity lines
        helicity_lines = self.get_helicity_lines(matrix_element)
        replace_dict['helicity_lines'] = helicity_lines
    
        # Extract IC line
        ic_line = self.get_ic_line(matrix_element)
        replace_dict['ic_line'] = ic_line
    
        # Extract overall denominator
        # Averaging initial state color, spin, and identical FS particles
        den_factor_line = self.get_den_factor_line(matrix_element)
        replace_dict['den_factor_line'] = den_factor_line
    
        # Extract ngraphs
        ngraphs = matrix_element.get_number_of_amplitudes()
        replace_dict['ngraphs'] = ngraphs
    
        # Extract ncolor
        ncolor = max(1, len(matrix_element.get('color_basis')))
        replace_dict['ncolor'] = ncolor
    
        # Extract color data lines
        color_data_lines = self.get_color_data_lines(matrix_element)
        replace_dict['color_data_lines'] = "\n".join(color_data_lines)
    
        # Extract helas calls
        helas_calls = fortran_model.get_matrix_element_calls(\
                    matrix_element)
        replace_dict['helas_calls'] = "\n".join(helas_calls)
    
        # Extract nwavefuncs (important to place after get_matrix_element_calls
        # so that 'me_id' is set)
        nwavefuncs = matrix_element.get_number_of_wavefunctions()
        replace_dict['nwavefuncs'] = nwavefuncs
    
        # Extract amp2 lines
        amp2_lines = self.get_amp2_lines(matrix_element)
        replace_dict['amp2_lines'] = '\n'.join(amp2_lines)

        # Set the size of Wavefunction
        if not self.model or any([p.get('spin') in [4,5] for p in self.model.get('particles') if p]):
            replace_dict['wavefunctionsize'] = 20
        else:
            replace_dict['wavefunctionsize'] = 8
    
        # Extract JAMP lines
        jamp_lines = self.get_JAMP_lines(matrix_element)
    
        replace_dict['jamp_lines'] = '\n'.join(jamp_lines)
    
        realfile = open(os.path.join(self.template_path, \
                             'realmatrix_pwgamc.inc')).read()

        realfile = realfile % replace_dict
        
        # Write the file
        writer.writelines(realfile)
    
        return len([call for call in helas_calls if call.find('#') != 0]), ncolor, ngraphs


    def get_iflines_pdgs(self,pdg_list):
        return ".or.".join(['str.eq."%s"' % self.pdg2string(pdgs) \
                        for pdgs in pdg_list])


    def pdg2string(self, pdgs):
        """ converst a list of pdgs to a string
        in the powheg convention, the gluon has id=0
        """
        return ''.join(["%d" % p if p != 21 else "0" for p in pdgs])



    #===============================================================================
    # get_ij_lines
    #===============================================================================
    def get_ij_lines(self, fks_born):
        """returns the lines with the information on the particle number of the born 
        that splits"""
        info_list = fks_born.get_fks_info_list()
        lines = []
        if info_list:
            # if the reals have been generated, fill with the corresponding value of ij if
            # ij is massless, or with 0 if ij is massive (no collinear singularity)
            ij_list = [info['fks_info']['ij']if \
                    fks_born.born_matrix_element['processes'][0]['legs'][info['fks_info']['ij']-1]['massless'] \
                    else 0 for info in info_list]

            # note that now (for powheg) we restrict the list to contain only elements which are different
            ij_list = list(set(ij_list))

            lines.append('integer n_ij')
            lines.append('parameter(n_ij=%d)' % len(ij_list))
            lines.append('INTEGER IJ_VALUES(n_ij)')
            lines.append('DATA IJ_VALUES /' + ', '.join(['%d' % ij for ij in ij_list]) + '/')
        else:
            #otherwise just put the first leg
            lines.append('INTEGER IJ_VALUES(1)')
            lines.append('DATA IJ_VALUES / 1 /')

        return lines



    #===============================================================================
    # get_born_pdg_lines
    #===============================================================================
    def get_born_pdg_lines(self, fks_born):
        """returns the lines with the information on the pdg ids of the born"""
        lines = ['integer bpdg(nexternal-1)']
        lines.append('data bpdg / %s /' % ', '.join(['%d' % l['id'] for l in \
                fks_born.born_matrix_element['processes'][0]['legs']]))

        return lines



    # functions that return a tag for the matrix elements
    def get_real_tag(self, n):
        """returns a tag for real matrix elements
        """
        return "R%d" % n


    def get_born_tag(self, n):
        """returns a tag for real matrix elements
        """
        return "B%d" % n


    def get_virt_tag(self, n):
        """returns a tag for real matrix elements
        """
        return "V%d" % n



    def finalize(self, matrix_elements, history, mg5options, flaglist):
        """Finalize FKS directory by creating jpeg diagrams, html
                pages,proc_card_mg5.dat and madevent.tar.gz and create the MA5 card if
        necessary."""
        
        devnull = os.open(os.devnull, os.O_RDWR)
        try:
            res = misc.call([mg5options['lhapdf'], '--version'], \
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except Exception:
            res = 1
        if res != 0:
            logger.info('The value for lhapdf in the current configuration does not ' + \
                        'correspond to a valid executable.\nPlease set it correctly either in ' + \
                        'input/mg5_configuration or with "set lhapdf /path/to/lhapdf-config" ' + \
                        'and regenrate the process. \nTo avoid regeneration, edit the ' + \
                        ('%s/Cards/amcatnlo_configuration.txt file.\n' % self.dir_path ) + \
                        'Note that you can still compile and run aMC@NLO with the built-in PDFs\n')
        
        compiler_dict = {'fortran':  mg5options['fortran_compiler'],
                             'cpp':  mg5options['cpp_compiler'],
                             'f2py':  mg5options['f2py_compiler']}
        

        self.create_run_card(matrix_elements.get_processes(), history)

#        # write the model functions get_mass/width_from_id
        filename = os.path.join(self.dir_path,'Source','MODEL','get_mass_width_fcts.f')
        makeinc = os.path.join(self.dir_path,'Source','MODEL','makeinc.inc')
        self.write_get_mass_width_file(writers.FortranWriter(filename), makeinc, self.model)

#        # Write maxconfigs.inc based on max of ME's/subprocess groups

        filename = os.path.join(self.dir_path,'Source','maxconfigs.inc')
        self.write_maxconfigs_file(writers.FortranWriter(filename),
                                   matrix_elements.get_max_configs())
        
#        # Write maxparticles.inc based on max of ME's/subprocess groups
        filename = os.path.join(self.dir_path,'Source','maxparticles.inc')
        self.write_maxparticles_file(writers.FortranWriter(filename),
                                     matrix_elements.get_max_particles())

        # Check for compiler
        fcompiler_chosen = self.set_fortran_compiler(compiler_dict)
        ccompiler_chosen = self.set_cpp_compiler(compiler_dict['cpp'])

#
        if os.path.isdir('Cards'):
            output_file = os.path.join('Cards', 'proc_card_mg5.dat')
            history.write(output_file)

        # Duplicate run_card and FO_analyse_card
        for card in ['run_card', 'FO_analyse_card', 'shower_card']:
            try:
                shutil.copy(pjoin(self.dir_path, 'Cards',
                                         card + '.dat'),
                           pjoin(self.dir_path, 'Cards',
                                        card + '_default.dat'))
            except IOError:
                logger.warning("Failed to copy " + card + ".dat to default")

        # update the model makefile
        modelmake = open(pjoin(self.dir_path, 'Source', 'MODEL', 'makefile')).read()
        modelmake = modelmake.replace("../../bin/aMCatNLO treatcards param", \
                                      "../../bin/treatcards.py ../../Cards/param_card.dat")
        outmake = open(pjoin(self.dir_path, 'Source', 'MODEL', 'makefile'), 'w')
        outmake.write(modelmake)
        outmake.close()




    def write_makefile_TIR(self, writer, link_tir_libs,tir_libs,tir_include=[]):
        """ Create the file makefile_loop which links to the TIR libraries."""
            
        file = open(os.path.join(self.dir_path,
                                 'SubProcesses','makefile_loop.inc')).read()  
        replace_dict={}
        replace_dict['link_tir_libs']=' '.join(link_tir_libs)
        replace_dict['tir_libs']=' '.join(tir_libs)
        replace_dict['dotf']='%.f'
        replace_dict['doto']='%.o'
        replace_dict['tir_include']=' '.join(tir_include)
        file=file%replace_dict
        if writer:
            writer.writelines(file)
        else:
            return file


    def write_proc_label(self, fname, pdg_list):
        """ create the proc label file with the
        pdg ids of all processes
        """
        # pdg_list must be flattened

        lines = []
        for i, ids in enumerate(sum(pdg_list, [])):
            lines.append("%d   %d   %s" %  (i+1, len(ids),  " ".join(["%d" % id if id != 21 else "0" for id in ids]))) 

        fout = open(fname, 'w')
        fout.write('\n'.join(lines))
        fout.close()



    #===========================================================================
    # copy_python_files 
    #===========================================================================        
    def copy_python_files(self):
        """copy python files required for the Template
        at the moment only check_param_card"""

        files_to_copy = [ \
          pjoin('iolibs','file_writers.py'),
          pjoin('iolibs','files.py'),
          pjoin('various','misc.py'),
          pjoin('..','models','check_param_card.py'),
          pjoin('__init__.py')
        ]
        
        for cp_file in files_to_copy:
            cp(pjoin(export_fks._file_path,cp_file),
                pjoin(self.dir_path,'bin','internal',os.path.basename(cp_file)))
