###################################################
#                                                   #
#  Source file of the interface for the             #
#  MadSTR plugin of MG5aMC.                          #
#                                                   #
#####################################################

import os
import subprocess
import logging
import itertools
import sys
import random
import re
import time
import shutil
import signal
import multiprocessing
import six.moves.cPickle

from madgraph import MadGraph5Error, InvalidCmd, MG5DIR
import madgraph.interface.master_interface as master_interface
import madgraph.iolibs.helas_call_writers as helas_call_writers
import madgraph.fks.fks_helas_objects as fks_helas
import madgraph.iolibs.files as files

import MG5aMC_PWG.mg5amcpwg_exporter as pwgamc_exporter

plugin_path = os.path.dirname(os.path.realpath( __file__ ))

logger = logging.getLogger('MadSTR_plugin.Interface')

pjoin = os.path.join


class PwgAmcInterfaceError(MadGraph5Error):
    """ Error from the resummation interface. """

class PwgAmcInvalidCmd(InvalidCmd):
    """ Invalid command issued to the resummation plugin. """


class PwgAmcInterface(master_interface.MasterCmd):
    """ Interface for steering the Resummation tasks. """

    # Change the prompt 
    def preloop(self, *args, **opts):
        """only change the prompt after calling  the mother preloop command"""
        super(PwgAmcInterface, self).preloop(*args,**opts)
        # The colored prompt screws up the terminal for some reason.
        #self.prompt = '\033[92mPY8Kernels > \033[0m'
        self.prompt = 'MG5_aMC-PWG > '


    def check_powhegbox_path(self):
        """check that the PowhegBox path in the Powheg Makefile exists
        """
        makefile = pjoin(plugin_path, "mg5amcpwgTemplate", "TEMPLATE", "Makefile")
        lines = open(makefile).read().split('\n')
        found = False
        for l in lines:
            if l.startswith("BOX="):
                boxpath = l.split("BOX=")[1]
                found = True
                break
        if not found:
            raise MadGraph5Error("Invalid Powheg Makefile %s" % makefile) 

        return boxpath.strip()


    def do_set_powhegbox_path(self, line):
        """ replace the PowhegBox path in the Powheg Makefile"""
        path = line.strip()
        if not path:
            raise self.InvalidCmd("Usage:\n" + \
                    "set_powhegbox_path /PATH/TO/POWHEG-BOX-V2")
        elif not os.path.isdir(path):
            raise self.InvalidCmd("Invalid or not existent path:\n %s" % path)
        else:
            makefile = pjoin(plugin_path, "mg5amcpwgTemplate", "TEMPLATE", "Makefile")
            lines = open(makefile).read().split('\n')
            found = False
            for l in lines:
                if l.startswith("BOX="):
                    lines[lines.index(l)] = "BOX=%s" % path
                    found = True
                    break
            if not found:
                raise MadGraph5Error("Invalid Powheg Makefile %s" % makefile) 

            make_out = open(makefile, 'w')
            make_out.write('\n'.join(lines) + '\n')

            


    def do_generate(self, *args, **opts):
        # first check that the powhegbox path has been set
        powhegbox_path = self.check_powhegbox_path()
        if not powhegbox_path:
            raise self.InvalidCmd("No PowhegBox-V2 path found. Please set it with\n" + \
                    "set_powhegbox_path /PATH/TO/POWHEG-BOX-V2")
        elif not os.path.isdir(powhegbox_path):
            raise self.InvalidCmd("Invalid PowhegBox-V2 path:\n" + powhegbox_path + \
                    "n Please set the correct one with\n" + \
                    "set_powhegbox_path /PATH/TO/POWHEG-BOX-V2")
        else:
            logger.info("Using PowhegBox-V2 installation in %s" % powhegbox_path)

        # then generate as usual
        super(PwgAmcInterface, self).do_generate(*args,**opts)


    def do_add(self, *args, **opts):
        super(PwgAmcInterface, self).do_add(*args,**opts)


    def do_launch(self, *args, **opts):
        raise self.InvalidCmd('The launch command is not available for MG5aMC-PWG.\n' + 
                              'Please run from inside the output folder (see the README therein).')



    def do_output(self, line):
        """output command: if no os divergences are there or if LO run has
        been generated nothing has to be done.
        Otherwise, the diagrams for the on-shell resonances 
        need to be exported too
        """
        if not hasattr(self, '_fks_multi_proc') or not self._fks_multi_proc:
            #MZMZ in these cases we should also switch the interface
            # or we just do no output
            super(PwgAmcInterface, self).do_output(line)
            return

        args = self.split_arg(line)
        # Check Argument validity
        self.check_output(args)
        
        noclean = '-noclean' in args
        force = '-f' in args 
        nojpeg = '-nojpeg' in args
        main_file_name = ""
        try:
            main_file_name = args[args.index('-name') + 1]
        except Exception:
            pass

        # For NLO, the group_subprocesses is automatically set to false
        group_processes = False

        # MZ these are copied from the V4Factory...
        # It may not be the most efficient way
        # ==========================================================================
        # First treat the MadLoop5 standalone case       
        MadLoop_SA_options = {'clean': not noclean, 
          'complex_mass':self.options['complex_mass_scheme'],
          'export_format':'madloop', 
          'mp':True,
          'loop_dir': os.path.join(self._mgme_dir,'Template','loop_material'),
          'cuttools_dir': self._cuttools_dir,
          'iregi_dir':self._iregi_dir,
          'golem_dir':self.options['golem'],
          'samurai_dir':self.options['samurai'],
          'ninja_dir':self.options['ninja'],
          'collier_dir':self.options['collier'],
          'fortran_compiler':self.options['fortran_compiler'],
          'f2py_compiler':self.options['f2py_compiler'],
          'output_dependencies':self.options['output_dependencies'],
          'SubProc_prefix':'P',
          'compute_color_flows':self.options['loop_color_flows'],
          'mode': 'reweight' if self._export_format == "standalone_rw" else '',
          'cluster_local_path': self.options['cluster_local_path'],
          'output_options': {'group_subprocesses': False}
          }
        # initialize the writer

        if self._export_format in ['NLO']:
            to_pass = dict(MadLoop_SA_options)
            to_pass['mp'] = len(self._fks_multi_proc.get_virt_amplitudes()) > 0
            to_pass['export_format'] = 'FKS5_optimized'
            self._curr_exporter = pwgamc_exporter.PwgAmcExporter(self._export_dir, to_pass)
            
            self._curr_exporter.pass_information_from_cmd(self)

        # check if a dir with the same name already exists
        if not force and not noclean and os.path.isdir(self._export_dir)\
               and self._export_format in ['NLO']:
            # Don't ask if user already specified force or noclean
            logger.info('INFO: directory %s already exists.' % self._export_dir)
            logger.info('If you continue this directory will be deleted and replaced.')
            answer = self.ask('Do you want to continue?', 'y', ['y','n'], 
                                                timeout=self.options['timeout'])
            if answer != 'y':
                raise self.InvalidCmd('Stopped by user request')

        # if one gets here either used -f or answered yes to the question about
        # removing the dir
        if os.path.exists(self._export_dir):
            shutil.rmtree(self._export_dir)

        # Make a Template Copy
        if self._export_format in ['NLO']:
            self._curr_exporter.copy_fkstemplate()

        # Reset _done_export, since we have new directory
        self._done_export = False

        # Perform export and finalize right away
        self.export(nojpeg, main_file_name, group_processes=group_processes)

        # Pass potential new information generated during the export.
        self._curr_exporter.pass_information_from_cmd(self)

        # Automatically run finalize
        self.finalize(nojpeg)
            
        # Generate the virtuals if from OLP
        if self.options['OLP']!='MadLoop':
            self._curr_exporter.generate_virtuals_from_OLP(
              self.born_processes_for_olp,self._export_dir,self.options['OLP'])
                
        # Remember that we have done export
        self._done_export = (self._export_dir, self._export_format)

        # Reset _export_dir, so we don't overwrite by mistake later
        self._export_dir = None



    # Export a matrix element  
    def export(self, nojpeg = False, main_file_name = "", group_processes=False):
        """Export a generated amplitude to file"""

        self._curr_helas_model = helas_call_writers.FortranUFOHelasCallWriter(self._curr_model)
        def generate_matrix_elements(self, group=False):
            """Helper function to generate the matrix elements before
            exporting"""

            # Sort amplitudes according to number of diagrams,
            # to get most efficient multichannel output
            self._curr_amps.sort(key = lambda a: a.get_number_of_diagrams(), reverse=True)

            cpu_time1 = time.time()
            ndiags = 0
            if not self._curr_matrix_elements.get_matrix_elements():
                if group:
                    raise MadGraph5Error("Cannot group subprocesses when "+\
                                                              "exporting to NLO")
                else:
                    self._curr_matrix_elements = \
                             fks_helas.FKSHelasMultiProcess(\
                                self._fks_multi_proc, 
                                loop_optimized= self.options['loop_optimized_output'])
                    
                    if not self.options['low_mem_multicore_nlo_generation']: 
                        # generate the code the old way
                        ndiags = sum([len(me.get('diagrams')) for \
                                      me in self._curr_matrix_elements.\
                                      get_matrix_elements()])
                        # assign a unique id number to all process and
                        # generate a list of possible PDF combinations
                        uid = 0 
                        initial_states=[]
                        for me in self._curr_matrix_elements.get_matrix_elements():
                            uid += 1 # update the identification number
                            me.get('processes')[0].set('uid', uid)
                            try:
                                initial_states.append(sorted(list(set((p.get_initial_pdg(1),p.get_initial_pdg(2)) for \
                                                                      p in me.born_matrix_element.get('processes')))))
                            except IndexError:
                                initial_states.append(sorted(list(set((p.get_initial_pdg(1)) for \
                                                                      p in me.born_matrix_element.get('processes')))))
                        
                            for fksreal in me.real_processes:
                            # Pick out all initial state particles for the two beams
                                try:
                                    initial_states.append(sorted(list(set((p.get_initial_pdg(1),p.get_initial_pdg(2)) for \
                                                                 p in fksreal.matrix_element.get('processes')))))
                                except IndexError:
                                    initial_states.append(sorted(list(set((p.get_initial_pdg(1)) for \
                                                                 p in fksreal.matrix_element.get('processes')))))
                                    
                            
                        # remove doubles from the list
                        checked = []
                        for e in initial_states:
                            if e not in checked:
                                checked.append(e)
                        initial_states=checked

                        self._curr_matrix_elements.set('initial_states',initial_states)

                    else:
                        #new NLO generation
                        if self._curr_matrix_elements['has_loops']:
                            self._curr_exporter.opt['mp'] = True
                        self._curr_exporter.model = self._curr_model
                        ndiags = 0

            cpu_time2 = time.time()
            return ndiags, cpu_time2 - cpu_time1

        # Start of the actual routine

        ndiags, cpu_time = generate_matrix_elements(self, group=group_processes)

        path = self._export_dir

        if self._export_format in ['NLO']:
            path = os.path.join(path, 'SubProcesses')

            #_curr_matrix_element is a FKSHelasMultiProcess Object 
            self._fks_directories = []
            proc_charac = self._curr_exporter.proc_characteristic 
            for charac in ['has_isr', 'has_fsr', 'has_loops']:
                proc_charac[charac] = self._curr_matrix_elements[charac]

            # prepare for the generation
            # glob_directories_map is for the new NLO generation
            global glob_directories_map
            glob_directories_map = []

            # Save processes instances generated
            self.born_processes_for_olp = []
            self.born_processes = []
            calls = 0
            if not self.options['low_mem_multicore_nlo_generation']:
                calls, born_processes = self._curr_exporter.generate_pwgamc_files(\
                         self._curr_matrix_elements, \
                         self._curr_helas_model, \
                         path,self.options['OLP'])

                # flatten out the list of lists of born processes
                self.born_processes = sum(born_processes, [])
                # take the first process of each list for the OLPs
                self.born_processes_for_olp = [procs[0] for procs in born_processes]

            else:
                raise MadGraph5Error, "low_mem_multicore_nlo_generation not yet supported "+\
                                                              "with MG5aMC-PWG"

            if self.options['low_mem_multicore_nlo_generation']:
                raise MadGraph5Error, "low_mem_multicore_nlo_generation not yet supported "+\
                                                              "with MG5aMC-PWG"
                # start the pool instance with a signal instance to catch ctr+c
                logger.info('Writing directories...')
                original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
                if self.ncores_for_proc_gen < 0: # use all cores
                    pool = multiprocessing.Pool(maxtasksperchild=1)
                else:
                    pool = multiprocessing.Pool(processes=self.ncores_for_proc_gen,maxtasksperchild=1)
                signal.signal(signal.SIGINT, original_sigint_handler)
                try:
                    # the very large timeout passed to get is to be able to catch
                    # KeyboardInterrupts
                    diroutputmap = pool.map_async(generate_directories_fks_async,
                                                  range(len(glob_directories_map))).get(9999999)
                except KeyboardInterrupt:
                    pool.terminate()
                    raise KeyboardInterrupt 
    
                pool.close()
                pool.join()
                
                #clean up tmp files containing final matrix elements
                for mefile in self._curr_matrix_elements.get('matrix_elements'):
                    os.remove(mefile)

                for charac in ['nexternal', 'ninitial']:
                    proc_charac[charac] = self._curr_exporter.proc_characteristic[charac]
                # ninitial and nexternal
                proc_charac['nexternal'] = max([diroutput[4] for diroutput in diroutputmap])
                ninitial_set = set([diroutput[3] for diroutput in diroutputmap])
                if len(ninitial_set) != 1:
                    raise MadGraph5Error("Invalid ninitial values: %s" % ' ,'.join(list(ninitial_set)))    
                proc_charac['ninitial'] = list(ninitial_set)[0]

                self.born_processes = []
                self.born_processes_for_olp = []
                max_loop_vertex_ranks = []
                
                for diroutput in diroutputmap:
                    calls = calls + diroutput[0]
                    self._fks_directories.extend(diroutput[1])
                    max_loop_vertex_ranks.append(diroutput[2])
                    self.born_processes.extend(diroutput[5])
                    self.born_processes_for_olp.append(diroutput[5][0])

            else:
                max_loop_vertex_ranks = [me.get_max_loop_vertex_rank() for \
                                         me in self._curr_matrix_elements.get_virt_matrix_elements()]

            card_path = os.path.join(path, os.path.pardir, 'SubProcesses', \
                                     'procdef_mg5.dat')
            
            if self.options['loop_optimized_output'] and \
               len(max_loop_vertex_ranks) > 0:
                self._curr_exporter.write_coef_specs_file(max_loop_vertex_ranks)
            ##if self._generate_info:
            ##    self._curr_exporter.write_procdef_mg5(card_path, #
            ##                    self._curr_model['name'],
            ##                    self._generate_info)
                try:
                    cmd.Cmd.onecmd(self, 'history .')
                except Exception:
                    logger.debug('fail to run command \"history cmd\"')
                    pass
            subproc_path = os.path.join(path, os.path.pardir, 'SubProcesses', \
                                     'initial_states_map.dat')
            ##self._curr_exporter.write_init_map(subproc_path,
            ##                    self._curr_matrix_elements.get('initial_states'))
            
        cpu_time1 = time.time()


    def finalize(self, *args, **opts):
        """wraps the finalize funcion of the madgraph_interface, and
        prints some extra informations
        """

        # swhich the MG5_interface logger off
        logger_off = logging.getLogger('cmdprint')
        old_level = logger_off.level
        logger_off.setLevel(logging.WARNING)
        # call the finalize function of the MG5_interface
        super(PwgAmcInterface, self).finalize(*args,**opts)
        # restore the level
        logger_off.setLevel(old_level)

        # finally copy README/LICENSE and print last info
        files.cp(pjoin(plugin_path, 'README'), self._export_dir)
        files.cp(pjoin(plugin_path, 'LICENSE'), self._export_dir)
        logger.info('\n\n' + 'Please see\n' + self._export_dir + '/README\n' + \
                'for instructions on how to use and run the code.\n')

            
